import pyark.cva_client as cva_client
from protocols.protocol_7_4.cva import VariantsCoordinates, VariantCoordinates


class LiftOverClient(cva_client.CvaClient):

    _BASE_ENDPOINT = "lift-overs"

    def __init__(self, **params):
        cva_client.CvaClient.__init__(self, **params)
        self.variants_client = self.variants()

    def lift_over_by_identifiers(self, variant_identifiers, **params):
        """
        :param variant_identifiers: the list of variant identifiers in format
        {assembly}:{chromosome}:{position}:{reference}:{alternate}
        :type variant_identifiers: list
        :return list of VariantCoordinates
        :rtype: list
        """
        variant_coordinates_list = self.variants_client.variant_ids_to_coordinates(variant_identifiers)
        variants_coordinates = VariantsCoordinates()
        variants_coordinates.variants = variant_coordinates_list

        results, _ = self._post(self._BASE_ENDPOINT, payload=variants_coordinates.toJsonDict(), **params)
        assert len(results) == len(variant_identifiers), "Some variants failed to lift over"

        return [VariantCoordinates.fromJsonDict(x) for x in results]

    def lift_over_by_variants_coordinates(self, variant_coordinates_list, **params):
        """
        :param variant_coordinates_list: the list of VariantCoordinates
        :type variant_coordinates_list: list
        :return list of VariantCoordinates
        :rtype: list
        """
        variants_coordinates = VariantsCoordinates()
        variants_coordinates.variants = variant_coordinates_list

        results, _ = self._post(self._BASE_ENDPOINT, payload=variants_coordinates.toJsonDict(), **params)
        assert len(results) == len(variant_coordinates_list), "Some variants failed to lift over"

        return [VariantCoordinates.fromJsonDict(x) for x in results]



import logging
import requests
import datetime
import abc
from furl import furl
import pyark.backoff_retrier as backoff_retrier
from pyark.errors import CvaServerError, CvaClientError


class RestClient(object):

    _session = requests.Session()

    def __init__(self, url_base, endpoint_base=None, retries=5):
        self._url_base = url_base
        self._endpoint_base = endpoint_base
        self._headers = {
            'Accept': 'application/json'
        }
        self._token = None
        self._renewed_token = False
        # decorates the REST verbs with retries
        self._get = backoff_retrier.wrapper(self._get, retries)
        self._post = backoff_retrier.wrapper(self._post, retries)
        self._delete = backoff_retrier.wrapper(self._delete, retries)

    def _build_url(self, endpoint):
        f = furl(self._url_base)
        segments = []
        if self._endpoint_base:
            segments = self._endpoint_base.split("/")
        if isinstance(endpoint, (str,)):
            endpoint = endpoint.split("/")
        if isinstance(endpoint, (list,)):
            segments = segments + endpoint
        else:
            segments.append(endpoint)
        f.path.segments = segments
        return f.url

    def _set_authenticated_header(self, renew_token=False):
        if not self._token or renew_token:
            self._token = self._get_token()
        self._headers["Authorization"] = "{token}".format(token=self._token)

    @abc.abstractmethod
    def _get_token(self):
        raise ValueError("Not implemented")

    def _post(self, endpoint, payload, session=True, verify=True, **params):
        if endpoint is None or payload is None:
            raise ValueError("Must define payload and endpoint before post")
        url = self._build_url(endpoint)
        if session:
            response = self._session.post(url, json=payload, params=params, headers=self._headers)
        else:
            response = requests.post(url, json=payload, params=params, headers=self._headers)
        request = "{method} {url}".format(
            method="POST", url="{}?{}".format(url, "&".join(RestClient._build_parameters(params))))
        logging.info(request)
        if verify:
            self._verify_response(response, request)
        return response.json(), dict(response.headers)

    def _get(self, endpoint, session=True, **params):
        if endpoint is None:
            raise ValueError("Must define endpoint before get")
        url = self._build_url(endpoint)
        if session:
            response = self._session.get(url, params=params, headers=self._headers)
        else:
            response = requests.get(url, params=params, headers=self._headers)
        request = "{method} {url}".format(
            method="GET", url="{}?{}".format(url, "&".join(RestClient._build_parameters(params))))
        logging.info(request)
        self._verify_response(response, request)
        return response.json(), dict(response.headers)

    def _patch(self, endpoint, session=True, **params):
        if endpoint is None:
            raise ValueError("Must define endpoint before patch")
        url = self._build_url(endpoint)
        if session:
            response = self._session.patch(url, params=params, headers=self._headers)
        else:
            response = requests.patch(url, params=params, headers=self._headers)
        request = "{method} {url}".format(
            method="PATCH", url="{}?{}".format(url, "&".join(RestClient._build_parameters(params))))
        logging.info(request)
        self._verify_response(response, request)
        return response.json(), dict(response.headers)

    def _delete(self, endpoint, **params):
        if endpoint is None:
            raise ValueError("Must define endpoint before get")
        url = self._build_url(endpoint)
        response = self._session.delete(url, params=params, headers=self._headers)
        request = "{method} {url}".format(
            method="DELETE", url="{}?{}".format(url, "&".join(RestClient._build_parameters(params))))
        logging.info(request)
        self._verify_response(response, request)
        return response.json(), dict(response.headers)

    @staticmethod
    def _build_parameters(params):
        parsed_params = []
        for k, v in params.items():
            if isinstance(v, list):
                parsed_params.extend(["{}={}".format(k, e) for e in v])
            else:
                parsed_params.append("{}={}".format(k, v))
        return parsed_params

    def _verify_response(self, response, request):
        logging.debug("{date} response status code {status}".format(
            date=datetime.datetime.now(),
            status=response.status_code)
        )
        if response.status_code != 200:
            # first 403 renews the token, second 403 in a row fails
            if response.status_code in (403, 401) and not self._renewed_token:
                # renews the token if unauthorised
                self._set_authenticated_header(renew_token=True)
                self._renewed_token = True
                # RequestException will trigger a retry and with the renewed token it may work
                self.log_error(response, request)
                raise requests.exceptions.RequestException(response=response)
            # ValueError will not
            if 500 <= response.status_code < 600:
                self.log_error(response, request)
                raise CvaServerError("{}:{}".format(response.status_code, response.text))
            elif 400 <= response.status_code < 500:
                if response.status_code != 404:     # we want to hide 404 for empty results to the end user
                    self.log_error(response, request)
                    raise CvaClientError("{}:{}".format(response.status_code, response.text))
            else:
                self.log_error(response, request)
                raise ValueError("{}:{}".format(response.status_code, response.text))
        else:
            # once a 200 response token is not anymore just renewed, it can be renewed again if a 403 arrives
            self._renewed_token = False

    def log_error(self, response, request):
        logging.error(request)
        logging.error("{} - {}".format(response.status_code, response.text))



from pyark.errors import CvaClientError

from pyark import cva_client
from protocols.protocol_7_4.cva import Transaction


class TransactionsClient(cva_client.CvaClient):

    _BASE_ENDPOINT = "transactions"

    def __init__(self, **params):
        cva_client.CvaClient.__init__(self, **params)

    def get_transaction(self, transaction_id):
        """
        :type transaction_id: str
        :rtype: Transaction
        """
        results, _ = self._get("{endpoint}/{identifier}".format(
            endpoint=self._BASE_ENDPOINT, identifier=transaction_id))
        result = self._render_single_result(results, as_data_frame=False)
        return Transaction.fromJsonDict(result) if result else None

    def retry_transaction(self, transaction_id):
        """
        :type transaction_id: str
        :rtype: Transaction
        """
        results, _ = self._patch("{endpoint}/{identifier}".format(
            endpoint=self._BASE_ENDPOINT, identifier=transaction_id))
        result = self._render_single_result(results, as_data_frame=False)
        return Transaction.fromJsonDict(result) if result else None

    def delete_transaction(self, **params):
        id = params.get('id', None)

        message = params.get('message', None)
        if not id:
            raise CvaClientError("You must specify an transaction id to delete a transaction")

        return self._delete(
            "{endpoint}/{identifier}".format(endpoint=self._BASE_ENDPOINT, identifier=params.get('id')),
            params={'message': message}
        )



import requests
import logging
import time
import random
from future.standard_library import install_aliases
install_aliases()
import urllib.error


def wrapper(func, retries):
    """
    This wrapper implements a truncated binary exponential backoff algorithm between retries.
    (https://en.wikipedia.org/wiki/Exponential_backoff#Binary_exponential_backoff)
    It only captures exceptions raised by the packages requests and urllib:
    * requests.exceptions.RequestException
    * requests.exceptions.ConnectionError
    * urllib.error.URLError
    Other exceptions will override any retries.

    :param func:       the wrapped function
    :param retries:    the maximum number of retries. -1 are infinite retries
    :return:           the return of the wrapped function if any
    """

    def retry(*args, **kwargs):

        results = None
        retries_count = 0
        backoff_iteration = 1
        truncate_iteration = 8
        success = False
        while not success:
            try:
                results = func(*args, **kwargs)
                success = True
            except (requests.exceptions.RequestException, requests.exceptions.ConnectionError, urllib.error.URLError) as ex:
                logging.error(str(ex))
                # retries a fixed number of times
                if retries != -1 and retries_count >= retries:
                    raise ex
                retries_count += 1
                # waits for an increasing random time
                random_sleep = random.randrange(0, min((2 ** backoff_iteration) - 1, 10))
                logging.info("Retrying connection after %s seconds" % str(random_sleep))
                time.sleep(random_sleep)
                # when it reaches the maximum value that it may wait it stops increasing time
                if backoff_iteration < truncate_iteration:
                    backoff_iteration += 1
        return results

    return retry



from enum import Enum
from protocols.protocol_7_4.cva import ReportEventEntry, Assembly, VariantCoordinates

import pyark.cva_client as cva_client
from pyark.models.wrappers import ReportEventEntryWrapper


class ReportEventsClient(cva_client.CvaClient):

    _BASE_ENDPOINT = "report-events"

    def __init__(self, **params):
        cva_client.CvaClient.__init__(self, **params)

    def count(self, **params):
        params['count'] = True
        return self.get_report_events(**params)

    def get_report_events(self, max_results=None, include_all=True, as_data_frame=False, **params):
        """
        :type as_data_frame: bool
        :type max_results: bool
        :type include_all: bool
        :type params: dict
        :rtype: generator
        """
        if params.get('count', False):
            results, next_page_params = self._get(self._BASE_ENDPOINT, **params)
            return results[0]
        else:
            if include_all:
                params['include'] = [self._INCLUDE_ALL]
            if not as_data_frame:
                def transformer(x): return ReportEventEntryWrapper.fromJsonDict(x)
            else:
                transformer = None
            return self._paginate(
                endpoint=self._BASE_ENDPOINT, max_results=max_results, as_data_frame=as_data_frame,
                transformer=transformer, **params)

    def get_variant_summary_by_ids(self, variant_ids, **params):
        """
        :type variant_ids: list
        :rtype: list
        """
        results, _ = self._post([self._BASE_ENDPOINT, "variant-summary-by-ids"], variant_ids, **params)
        return results
